import 'package:flutter/material.dart';
import 'package:pie_menu/src/pie_button.dart';
import 'package:pie_menu/src/pie_button_theme.dart';
import 'package:pie_menu/src/pie_canvas.dart';
import 'package:pie_menu/src/pie_menu.dart';
import 'package:pie_menu/src/pie_provider.dart';

/// Action display anchor point for the specified custom angle in [PieTheme].
enum PieAnchor { start, center, end }

/// Decides how to display the translucent canvas overlay.
enum PieOverlayStyle {
  /// Displays the overlay to cover the entire canvas,
  /// and re-renders the menu child on top of the overlay.
  ///
  /// This is the recommended style if your menu child is stateless.
  behind,

  /// Draws the overlay around the menu child using [CustomPainter].
  ///
  /// Use this style if you want to preserve the state of your menu child.
  /// You might experience some rendering issues when the menu is partially
  /// obscured by other widgets.
  around;
}

/// Defines the behavior and the appearance
/// of [PieCanvas] and [PieMenu] widgets.
class PieTheme {
  /// Creates a [PieTheme] to configure [PieMenu]s.
  const PieTheme({
    this.brightness = Brightness.light,
    this.overlayColor,
    this.pointerColor,
    this.pointerDecoration,
    this.buttonTheme = const PieButtonTheme(
      backgroundColor: Colors.blue,
      iconColor: Colors.white,
    ),
    this.buttonThemeHovered = const PieButtonTheme(
      backgroundColor: Colors.green,
      iconColor: Colors.white,
    ),
    this.iconSize,
    this.radius = 96,
    this.spacing = 6,
    this.customAngleDiff,
    this.angleOffset = 0,
    this.customAngle,
    this.customAngleAnchor = PieAnchor.center,
    this.buttonSize = 56,
    this.pointerSize = 40,
    this.tooltipPadding = const EdgeInsets.all(32),
    this.tooltipTextStyle,
    this.tooltipTextAlign,
    this.tooltipCanvasAlignment,
    this.tooltipUseFittedBox = false,
    this.pieBounceDuration = const Duration(seconds: 1),
    this.childBounceEnabled = true,
    this.childTiltEnabled = true,
    this.childBounceDuration = const Duration(milliseconds: 150),
    this.childBounceFactor = 0.95,
    this.childBounceCurve = Curves.easeOutCubic,
    this.childBounceReverseCurve = Curves.easeInCubic,
    this.fadeDuration = const Duration(milliseconds: 250),
    this.hoverDuration = const Duration(milliseconds: 250),
    this.delayDuration = const Duration(milliseconds: 350),
    this.leftClickShowsMenu = true,
    this.rightClickShowsMenu = false,
    this.overlayStyle = PieOverlayStyle.behind,
    this.childOpacityOnButtonHover = 0.5,
  });

  /// How the background and tooltip widgets should be displayed
  /// if they are not specified explicitly.
  final Brightness brightness;

  /// Preferably a translucent color for [PieCanvas] to display
  /// under the menu child, and on top of the other widgets.
  final Color? overlayColor;

  /// Custom color for the widget displayed in the center of [PieMenu].
  final Color? pointerColor;

  /// Decoration for the widget displayed in the center of [PieMenu].
  ///
  /// If specified, [pointerColor] will be ignored.
  final Decoration? pointerDecoration;

  /// Theme of [PieButton].
  final PieButtonTheme buttonTheme;

  /// Theme of [PieButton] when it is hovered.
  final PieButtonTheme buttonThemeHovered;

  /// Size of the icon to be displayed on the [PieButton].
  final double? iconSize;

  /// Distance between the [PieButton] and the center of [PieMenu].
  final double radius;

  /// Spacing between the [PieButton]s.
  final double spacing;

  /// Angle difference between the [PieButton]s in degrees.
  ///
  /// If specified, [spacing] will be ignored.
  final double? customAngleDiff;

  /// Angle offset in degrees for the actions.
  final double angleOffset;

  /// Display the menu actions in a specific angle in degrees.
  final double? customAngle;

  /// Action display alignment for the specified [customAngle].
  final PieAnchor customAngleAnchor;

  /// Size of [PieButton] circle.
  final double buttonSize;

  /// Size of the widget displayed in the center of [PieMenu].
  final double pointerSize;

  /// Padding value of the tooltip at the edges of [PieCanvas].
  final EdgeInsets tooltipPadding;

  /// Default text style for the tooltip widget.
  final TextStyle? tooltipTextStyle;

  /// Text alignment of the tooltip widget.
  final TextAlign? tooltipTextAlign;

  /// Alignment of the tooltip in the [PieCanvas].
  ///
  /// Setting this property will disable dynamic tooltip positioning.
  final Alignment? tooltipCanvasAlignment;

  /// Whether to wrap the tooltip with [FittedBox] widget.
  ///
  /// Can be used to display long tooltip texts in a single line.
  final bool tooltipUseFittedBox;

  /// Duration of [PieButton] bounce animation.
  final Duration pieBounceDuration;

  /// Whether to bounce the [PieMenu] child on press.
  final bool childBounceEnabled;

  /// Whether to tilt the [PieMenu] child on press.
  ///
  /// Only works if [childBounceEnabled] is set to true.
  final bool childTiltEnabled;

  /// Duration of menu child bounce animation.
  final Duration childBounceDuration;

  /// Distance of menu child bounce animation.
  final double childBounceFactor;

  /// Curve for the menu child bounce animation.
  final Curve childBounceCurve;

  /// Reverse curve for the menu child bounce animation.
  final Curve? childBounceReverseCurve;

  /// Duration of [PieMenu] fade animation.
  final Duration fadeDuration;

  /// Duration of [PieButton] hover animation.
  final Duration hoverDuration;

  /// Long press duration for [PieMenu] to display.
  ///
  /// Can be set to [Duration.zero] to display the menu immediately on tap.
  final Duration delayDuration;

  /// Whether to display the menu on left mouse click.
  final bool leftClickShowsMenu;

  /// Whether to display the menu on right mouse click.
  final bool rightClickShowsMenu;

  /// Decides how to display the translucent canvas overlay.
  ///
  /// [PieOverlayStyle.behind] is the recommended style
  /// if your menu child is stateless.
  ///
  /// Use [PieOverlayStyle.around] if you want to preserve the state of your
  /// menu child. However, you might experience some rendering issues
  /// when the menu is partially obscured by other widgets.
  final PieOverlayStyle overlayStyle;

  /// Opacity of the menu child when a button is hovered.
  final double childOpacityOnButtonHover;

  /// Displacement distance of [PieButton]s when hovered.
  double get hoverDisplacement => buttonSize / 8;

  Color get effectiveOverlayColor {
    return overlayColor ??
        (brightness == Brightness.light
            ? Colors.white.withOpacity(0.8)
            : Colors.black.withOpacity(0.8));
  }

  /// Returns the [PieTheme] defined in the closest [PieCanvas] instance
  /// that encloses the given context.
  static PieTheme of(BuildContext context) {
    return PieNotifier.of(context).canvasTheme;
  }

  /// Creates a copy of this theme but with the
  /// given fields replaced with the new values.
  PieTheme copyWith({
    Brightness? brightness,
    Color? overlayColor,
    Color? pointerColor,
    Decoration? pointerDecoration,
    PieButtonTheme? buttonTheme,
    PieButtonTheme? buttonThemeHovered,
    double? iconSize,
    double? radius,
    double? spacing,
    double? customAngleDiff,
    double? angleOffset,
    double? customAngle,
    PieAnchor? customAngleAnchor,
    double? buttonSize,
    double? pointerSize,
    EdgeInsets? tooltipPadding,
    TextStyle? tooltipTextStyle,
    TextAlign? tooltipTextAlign,
    Alignment? tooltipCanvasAlignment,
    bool? tooltipUseFittedBox,
    Duration? pieBounceDuration,
    bool? childBounceEnabled,
    bool? childTiltEnabled,
    Duration? childBounceDuration,
    double? childBounceDistance,
    Curve? childBounceCurve,
    Curve? childBounceReverseCurve,
    Duration? fadeDuration,
    Duration? hoverDuration,
    Duration? delayDuration,
    bool? leftClickShowsMenu,
    bool? rightClickShowsMenu,
    PieOverlayStyle? overlayStyle,
    double? childOpacityOnButtonHover,
  }) {
    return PieTheme(
      brightness: brightness ?? this.brightness,
      overlayColor: overlayColor ?? this.overlayColor,
      pointerColor: pointerColor ?? this.pointerColor,
      pointerDecoration: pointerDecoration ?? this.pointerDecoration,
      buttonTheme: buttonTheme ?? this.buttonTheme,
      buttonThemeHovered: buttonThemeHovered ?? this.buttonThemeHovered,
      iconSize: iconSize ?? this.iconSize,
      radius: radius ?? this.radius,
      spacing: spacing ?? this.spacing,
      customAngleDiff: customAngleDiff ?? this.customAngleDiff,
      angleOffset: angleOffset ?? this.angleOffset,
      customAngle: customAngle ?? this.customAngle,
      customAngleAnchor: customAngleAnchor ?? this.customAngleAnchor,
      buttonSize: buttonSize ?? this.buttonSize,
      pointerSize: pointerSize ?? this.pointerSize,
      tooltipPadding: tooltipPadding ?? this.tooltipPadding,
      tooltipTextStyle: tooltipTextStyle ?? this.tooltipTextStyle,
      tooltipTextAlign: tooltipTextAlign ?? this.tooltipTextAlign,
      tooltipCanvasAlignment:
          tooltipCanvasAlignment ?? this.tooltipCanvasAlignment,
      tooltipUseFittedBox: tooltipUseFittedBox ?? this.tooltipUseFittedBox,
      pieBounceDuration: pieBounceDuration ?? this.pieBounceDuration,
      childBounceEnabled: childBounceEnabled ?? this.childBounceEnabled,
      childTiltEnabled: childTiltEnabled ?? this.childTiltEnabled,
      childBounceDuration: childBounceDuration ?? this.childBounceDuration,
      childBounceFactor: childBounceDistance ?? childBounceFactor,
      childBounceCurve: childBounceCurve ?? this.childBounceCurve,
      childBounceReverseCurve:
          childBounceReverseCurve ?? this.childBounceReverseCurve,
      fadeDuration: fadeDuration ?? this.fadeDuration,
      hoverDuration: hoverDuration ?? this.hoverDuration,
      delayDuration: delayDuration ?? this.delayDuration,
      leftClickShowsMenu: leftClickShowsMenu ?? this.leftClickShowsMenu,
      rightClickShowsMenu: rightClickShowsMenu ?? this.rightClickShowsMenu,
      overlayStyle: overlayStyle ?? this.overlayStyle,
      childOpacityOnButtonHover:
          childOpacityOnButtonHover ?? this.childOpacityOnButtonHover,
    );
  }
}
